import numpy as np
from typing import List, Union, Tuple, Optional
from scipy.linalg import svd
import math

def get_plane_from_list_of_three_points(points: List[List[float]]) -> Union[np.ndarray, None]:
    """
    Get plane in form Ax + By + Cz + D = 0 from list of three points.

    :param points: Points.
    :type points: List[List[float]]
    :return: Plane.
    :rtype: Union[np.ndarray, None]

    :Example:

    ::

        >>> import coreransac
        >>> points = [[0, 0, 0], [1, 0, 0], [0, 1, 0]]
        >>> plane = customransac.get_plane_from_points(points)
        >>> plane
        array([0, 0, 1, 0])
    """
    p1 = np.array(points[0])
    p2 = np.array(points[1])
    p3 = np.array(points[2])
    v1 = p2 - p1
    v2 = p3 - p1
    if np.allclose(v1, np.zeros_like(v1)) or np.allclose(v2, np.zeros_like(v2)) or np.allclose(v1, v2):
        print("Points are collinear, cannot get plane.")
        return None
    normal = np.cross(v1, v2)
    d = -np.dot(normal, p1)
    A, B, C = normal
    return np.array([A, B, C, d])

def get_limits_of_graph_from_limits_of_object(min_x: float, max_x: float, min_y: float, max_y: float) -> Tuple[float, float, float, float]:
    """
    Get limits of graph from limits of object. The (0,0) point should be in the center of the graph and the object should be whole visible. 
    The visible zone should be square. This is useful for plotting.

    :param min_x: Minimum x.
    :type min_x: float
    :param max_x: Maximum x.
    :type max_x: float
    :param min_y: Minimum y.
    :type min_y: float
    :param max_y: Maximum y.
    :type max_y: float
    :return: Limits of graph.
    :rtype: Tuple[float, float, float, float]
    """
    limit = max(abs(min_x), abs(max_x), abs(min_y), abs(max_y))
    return (-limit, limit, -limit, limit)

def get_limits_of_3d_graph_from_limits_of_object(min_x: float, max_x: float, min_y: float, max_y: float, min_z: float, max_z: float) -> \
                                                Tuple[float, float, float, float, float, float]:
    """
    Get limits of graph from limits of object. The (0,0,0) point should be in the center of the graph and the object should be whole visible. 
    The visible zone should be square. This is useful for plotting.

    :param min_x: Minimum x.
    :type min_x: float
    :param max_x: Maximum x.
    :type max_x: float
    :param min_y: Minimum y.
    :type min_y: float
    :param max_y: Maximum y.
    :type max_y: float
    :param min_z: Minimum z.
    :type min_z: float
    :return: Limits of graph.
    :rtype: Tuple[float, float, float, float, float, float]
    """
    limit = max(abs(min_x), abs(max_x), abs(min_y), abs(max_y), abs(min_z), abs(max_z))
    return (-limit, limit, -limit, limit, -limit, limit)

def get_parallelogram_2d_vertices(center: List[float], normal1: List[float], normal2: List[float], length1: float, length2: float):
    '''
    Get vertices of parallelogram, given center, normal vectors, and lengths.

    :param center: Center.
    :type center: List[float]
    :param normal1: Normal vector 1.
    :type normal1: List[float]
    :param normal2: Normal vector 2.
    :type normal2: List[float]
    :param length1: Length 1.
    :type length1: float
    :param length2: Length 2.
    :type length2: float
    :return: Vertices.
    :rtype: List[List[float]]

    :Example:

    ::
    '''
    center = np.array(center)
    normal1 = np.array(normal1)
    normal2 = np.array(normal2)
    vertex1 = center + normal1 * length1 / 2 + normal2 * length2 / 2
    vertex2 = center - normal1 * length1 / 2 + normal2 * length2 / 2
    vertex3 = center - normal1 * length1 / 2 - normal2 * length2 / 2
    vertex4 = center + normal1 * length1 / 2 - normal2 * length2 / 2
    return [vertex1, vertex2, vertex3, vertex4]

def get_parallelepiped_3d_vertices(center: List[float], normal1: List[float], normal2: List[float], normal3: List[float], 
                                  length1: float, length2: float, length3: float)-> List[List[float]]:
    '''
    Get vertices of parallelepiped, given center, normal vectors, and lengths.

    :param center: Center.
    :type center: List[float]
    :param normal1: Normal vector 1.
    :type normal1: List[float]
    :param normal2: Normal vector 2.
    :type normal2: List[float]
    :param normal3: Normal vector 3.
    :type normal3: List[float]
    :param length1: Length 1.
    :type length1: float
    :param length2: Length 2.
    :type length2: float
    :param length3: Length 3.
    :type length3: float
    :return: Vertices.
    :rtype: List[List[float]]

    :Example:

    ::
    '''
    center = np.array(center)
    normal1 = np.array(normal1)
    normal2 = np.array(normal2)
    normal3 = np.array(normal3)
    vertex1 = center + normal1 * length1 / 2 + normal2 * length2 / 2 + normal3 * length3 / 2
    vertex2 = center - normal1 * length1 / 2 + normal2 * length2 / 2 + normal3 * length3 / 2
    vertex3 = center - normal1 * length1 / 2 - normal2 * length2 / 2 + normal3 * length3 / 2
    vertex4 = center + normal1 * length1 / 2 - normal2 * length2 / 2 + normal3 * length3 / 2

    vertex5 = center + normal1 * length1 / 2 + normal2 * length2 / 2 - normal3 * length3 / 2
    vertex6 = center - normal1 * length1 / 2 + normal2 * length2 / 2 - normal3 * length3 / 2
    vertex7 = center - normal1 * length1 / 2 - normal2 * length2 / 2 - normal3 * length3 / 2
    vertex8 = center + normal1 * length1 / 2 - normal2 * length2 / 2 - normal3 * length3 / 2
    return [vertex1, vertex2, vertex3, vertex4, vertex5, vertex6, vertex7, vertex8]

def get_parallelogram_3d_vertices(center: List[float], normal1: List[float], normal2: List[float],  
                                  length1: float, length2: float)-> List[List[float]]:
    '''
    Get vertices of parallelogram, given center, normal vectors, and lengths.

    :param center: Center.
    :type center: List[float]
    :param normal1: Normal vector 1.
    :type normal1: List[float]
    :param normal2: Normal vector 2.
    :type normal2: List[float]
    :param length1: Length 1.
    :type length1: float
    :param length2: Length 2.
    :type length2: float
    :return: Vertices.
    :rtype: List[List[float]]

    :Example:

    ::
    '''
    center = np.array(center)
    normal1 = np.array(normal1)
    normal2 = np.array(normal2)
    vertex1 = center + normal1 * length1 / 2 + normal2 * length2 / 2
    vertex2 = center - normal1 * length1 / 2 + normal2 * length2 / 2
    vertex3 = center - normal1 * length1 / 2 - normal2 * length2 / 2
    vertex4 = center + normal1 * length1 / 2 - normal2 * length2 / 2

    return [vertex1, vertex2, vertex3, vertex4]

def get_plane_equation(normal1, normal2, point):
    normal1 = np.array(normal1)
    normal2 = np.array(normal2)
    point = np.array(point)
    # Normalize the vectors
    normal1 = normal1 / np.linalg.norm(normal1)
    normal2 = normal2 / np.linalg.norm(normal2)

    # Calculate the normal vector of the plane
    normal = np.cross(normal1, normal2)
    normal = normal / np.linalg.norm(normal)

    # Calculate the value of D
    D = -np.dot(normal, point)

    # Multiply all coefficients by the reciprocal of D
    A, B, C = normal

    # Return the plane equation coefficients
    return A, B, C, D

def find_closest_plane(points: List[List[float]]) -> Tuple[float]:
    """
    Find the closest plane to a set of points based on Euclidean distance.

    :param points: Points.
    :type points: List[List[float]]
    :return: Plane equation coefficients.
    :rtype: Tuple[float]
    """
    # Center the points around the origin
    centroid = np.mean(points, axis=0)
    centered_points = points - centroid

    # Compute the singular value decomposition
    _, _, vh = svd(centered_points)

    # Extract the last row of V^T to get the normal vector of the plane
    plane_normal = vh[-1]

    # Normalize the normal vector
    plane_normal /= np.linalg.norm(plane_normal)

    # Choose a point on the plane as the centroid of the centered points
    plane_point = centroid

    # Compute the D coefficient of the plane equation
    D = -np.dot(plane_normal, plane_point)

    # Return the plane equation coefficients
    return (*plane_normal, D)

# compute the distance from a point to a plane
def get_distance_from_point_to_plane(point:Tuple[float, float, float], plane: Tuple[float, float, float, float])->float:
    '''
    Get the distance from a point to a plane.

    :param point: Point.
    :type point: Tuple[float, float, float]
    :param plane: Plane equation coefficients.
    :type plane: Tuple[float, float, float, float]
    :return: Distance from point to plane.
    :rtype: float

    :Example:

    ::
    '''
    A, B, C, D = plane
    x, y, z = point
    return abs(A * x + B * y + C * z + D) / math.sqrt(A * A + B * B + C * C)

# compute the distance from a np array of points to a plane
def get_distance_from_points_to_plane(points:np.ndarray, plane: Tuple[float, float, float, float])->np.ndarray:
    '''
    Get the distance from a np array of points to a plane.

    :param points: Points.
    :type points: np.ndarray
    :param plane: Plane equation coefficients.
    :type plane: Tuple[float, float, float, float]
    :return: Distance from points to plane.
    :rtype: np.ndarray

    :Example:

    ::
    '''
    A, B, C, D = plane
    x, y, z = points[:, 0], points[:, 1], points[:, 2]
    return abs(A * x + B * y + C * z + D) / math.sqrt(A * A + B * B + C * C)

def fit_plane_svd(points):
    centroid = np.mean(points, axis=0)
    shifted_points = points - centroid
    _, _, V = np.linalg.svd(shifted_points)
    normal = V[2]
    d = -np.dot(normal, centroid)
    # Compute sum of squared errors
    errors = np.abs(np.dot(points, normal) + d)
    sse = np.sum(errors ** 2)

    return normal[0], normal[1], normal[2], d, sse

def get_angle_between_vectors(v1, v2):
    v1 = np.array(v1)
    v2 = np.array(v2)
    return np.arccos(np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)))

def get_angle_between_lines(l1, l2):
    v1 = l1[1] - l1[0]
    v2 = l2[1] - l2[0]
    print("Line 1: ", l1)
    print("Line 2: ", l2)
    print("Vector 1: ", v1)
    print("Vector 2: ", v2)
    return np.arccos(np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)))

def get_intersection_points_of_line_with_cube(line: np.ndarray, cube_min: np.ndarray, cube_max: np.ndarray) -> np.ndarray:
    '''
    Get the intersection points of a line with a cube.

    :param line: Line described as two points.
    :type line: np.ndarray
    :param cube_min: Minimum point of the cube.
    :type cube_min: np.ndarray
    :param cube_max: Maximum point of the cube.
    :type cube_max: np.ndarray
    :return: Intersection points.
    :rtype: np.ndarray

    :Example:

    ::

        >>> import mrdja.geometry as geom
        >>> import mrdja.drawing as drawing
        >>> import numpy as np
        >>> line = np.array([[0, 0, 0], [1, 1, 1]])
        >>> cube_min = np.array([-2, -2, -1])
        >>> cube_max = np.array([1, 2, 2])
        >>> intersection_points = geom.get_intersection_points_of_line_with_cube(line, cube_min, cube_max)
        >>> intersection_points
        array([[ 1.,  1.,  1.],
              [-1., -1., -1.]])
    '''
    # Initialize an empty list to store intersection points.
    intersection_points = []

    # Define the planes representing the faces of the cube.
    # Each plane is defined as (A, B, C, D), where Ax + By + Cz + D = 0.
    # We'll calculate the six planes for the cube.
    
    # Plane equations for the cube faces (front, back, top, bottom, left, right).
    planes = [
        (1, 0, 0, -cube_min[0]),  # Front face
        (-1, 0, 0, cube_max[0]),  # Back face
        (0, 1, 0, -cube_min[1]),  # Top face
        (0, -1, 0, cube_max[1]),  # Bottom face
        (0, 0, 1, -cube_min[2]),  # Left face
        (0, 0, -1, cube_max[2])   # Right face
    ]

    # Check for intersection with each plane.
    for plane in planes:
        intersection_point = get_intersection_point_of_line_with_plane(line, plane)
        if intersection_point is not None:
            # Check if the intersection point is not already in the list.
            if not any(np.allclose(intersection_point, point) for point in intersection_points):
                # Check if the intersection point is within the bounds of the cube.
                if all(cube_min <= intersection_point) and all(intersection_point <= cube_max):
                    intersection_points.append(intersection_point)

    return np.array(intersection_points)


def get_intersection_point_of_line_with_plane(line: np.ndarray, plane: np.ndarray) -> Optional[np.ndarray]:
    '''
    Get the intersection point of a line with a plane. If it is parallel, return None.

    :param line: Line described as two points.
    :type line: np.ndarray
    :param plane: Plane described as Ax + By + Cz + D = 0.
    :type plane: np.ndarray
    :return: Intersection point.
    :rtype: Optional[np.ndarray]

    :Example:

    ::

        >>> import mrdja.geometry as geom
        >>> import numpy as np 
        >>> import mrdja.drawing as drawing
        >>> line = np.array([[0, 0, 0], [1, 1, 1]])
        >>> plane = np.array([0, 0, 1, -3])
        >>> intersection_point = geom.get_intersection_point_of_line_with_plane(line, plane)
        >>> intersection_point
        array([3., 3., 3.])
        >>> drawing.draw_line_extension_to_plane(line, plane)

    |drawing_draw_line_extension_to_plane_example|

    .. |drawing_draw_line_extension_to_plane_example| image:: ../../_static/images/drawing_draw_line_extension_to_plane_example.png

    '''
    # Step 1: Calculate the direction vector of the line.
    direction = line[1] - line[0]
    
    # Step 2: Calculate the normal vector of the plane.
    normal = plane[:3]
    
    # Step 3: Check if the line is parallel to the plane.
    dot_product = np.dot(direction, normal)
    if np.isclose(dot_product, 0, atol=1e-6):
        return None
    
    # Step 4: Calculate the parameter t for the line's equation.
    t = - (np.dot(normal, line[0]) + plane[3]) / dot_product
    
    # Step 5: Calculate the intersection point.
    intersection_point = line[0] + t * direction
    return intersection_point

def get_two_perpendicular_unit_vectors_in_plane(plane: np.ndarray) -> Tuple[np.ndarray, np.ndarray]: 
    '''
    Get two perpendicular unit vectors in a plane.

    :param plane: Plane described as Ax + By + Cz + D = 0.
    :type plane: np.ndarray
    :return: Two perpendicular unit vectors in the plane.
    :rtype: Tuple[np.ndarray, np.ndarray]

    :Example:

    ::

        >>> import mrdja.geometry as geom
        >>> import numpy as np
        >>> plane = np.array([0, 0, 1, -3])
        >>> perpendicular1, perpendicular2 = geom.get_two_perpendicular_unit_vectors_in_plane(plane)
        >>> perpendicular1
        array([0., 1., 0.])
        >>> perpendicular2
        array([-1., 0., 0.])
    '''
    # Step 1: Calculate the normal vector of the plane.
    normal = plane[:3]
    # Step 2: Calculate the first perpendicular vector.
    if normal[0] == 0 and normal[1] == 0:
        perpendicular1 = np.array([0, 1, 0])
    else:
        perpendicular1 = np.array([normal[1], -normal[0], 0])
    # Step 3: Calculate the second perpendicular vector.
    perpendicular2 = np.cross(normal, perpendicular1)
    # Step 4: Normalize the vectors.
    perpendicular1 = perpendicular1 / np.linalg.norm(perpendicular1)
    perpendicular2 = perpendicular2 / np.linalg.norm(perpendicular2)
    return perpendicular1, perpendicular2

def get_best_plane_from_points_from_two_segments(segment_1: np.ndarray, segment_2: np.ndarray) -> Tuple[np.ndarray, float]:
    '''
    Computes the best fitting plane to the four points of two segments.

    :param segment_1: The first segment.
    :type segment_1: np.ndarray
    :param segment_2: The second segment.
    :type segment_2: np.ndarray
    :return: The best fitting plane and the sum of squared errors.
    :rtype: Tuple[np.ndarray, float]

    :Example:

    ::

        >>> import mrdja.geometry as geom
        >>> import numpy as np
        >>> segment_1 = np.array([[0, 0, 0], [1, 0, 0]])
        >>> segment_2 = np.array([[0, 1, 0], [1, 1, 0]])
        >>> geom.get_best_plane_from_points_from_two_segments(segment_1, segment_2)
        (array([ 0.,  0.,  1., -0.]), 0.0)
        >>> # another example
        >>> segment_1 = np.array([[1, 2, 3], [4, 5, 6]])
        >>> segment_2 = np.array([[7, 8, 9], [10, 11, 12]])
        >>> geom.get_best_plane_from_points_from_two_segments(segment_1, segment_2)
        (array([ 0.81649658, -0.40824829, -0.40824829,  1.22474487]),
        1.0107280348144214e-29)
        >>> # another example 
        >>> segment_1 = np.array([[0, 0, 0], [1, 0, 0]])
        >>> segment_2 = np.array([[0, 1, 0], [1, 1, 1]])
        >>> geom.get_best_plane_from_points_from_two_segments(segment_1, segment_2)
        (array([-0.45440135, -0.45440135,  0.76618459,  0.2628552 ]),
        0.15692966918274637)

    '''
    points = np.array([segment_1[0], segment_1[1], segment_2[0], segment_2[1]])
    a, b, c, d, sse = fit_plane_svd(points)
    best_plane = np.array([a, b, c, d])
    return best_plane, sse

def get_point_of_plane_closest_to_given_point(plane: np.ndarray, point: np.ndarray) -> np.ndarray:
    '''
    Get the point of a plane closest to a given point.

    :param plane: Plane described as Ax + By + Cz + D = 0.
    :type plane: np.ndarray
    :param point: Point.
    :type point: np.ndarray
    :return: Point of plane closest to given point.
    :rtype: np.ndarray

    :Example:
    '''
    # Step 1: Calculate the normal vector of the plane.
    normal = plane[:3]
    # Step 2: Calculate the point of the plane closest to the given point.
    t = - (np.dot(normal, point) + plane[3]) / np.dot(normal, normal)
    closest_point = point + t * normal
    return closest_point

def get_centroid_of_points(points: np.ndarray) -> np.ndarray:
    '''
    Get the centroid of a set of points.

    :param points: Points.
    :type points: np.ndarray
    :return: Centroid.
    :rtype: np.ndarray

    :Example:

    ::

        >>> import mrdja.geometry as geom
        >>> import numpy as np
        >>> points = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])
        >>> centroid = geom.get_centroid_of_points(points)
        >>> centroid
        array([0.33333333, 0.33333333, 0.        ])

    '''
    return np.mean(points, axis=0)

def get_a_polygon_from_plane_equation_and_point(plane: np.ndarray, point: np.ndarray, scale: float = 1.0)->np.ndarray:
    '''
    Get a polygon from plane equation and point.

    :param plane: Plane equation coefficients.
    :type plane: np.ndarray
    :param point: Point.
    :type point: np.ndarray
    :param scale: Scale.
    :type scale: float
    :return: Polygon.
    :rtype: np.ndarray

    :Example:

    ::

        >>> import mrdja.geometry as geom
        >>> import numpy as np
        >>> plane = np.array([0, 0, 1, -3])
        >>> point = np.array([1, 1, 1])
        >>> polygon = geom.get_a_polygon_from_plane_equation_and_point(plane, point)
        >>> polygon
    '''
    # Step 1: Calculate the normal vector of the plane.
    normal = plane[:3]
    # Step 2: Calculate the point of the plane closest to the given point.
    t = - (np.dot(normal, point) + plane[3]) / np.dot(normal, normal)
    closest_point = point + t * normal
    # Step 3: Calculate two perpendicular unit vectors in the plane.
    perpendicular1, perpendicular2 = get_two_perpendicular_unit_vectors_in_plane(plane)
    # Step 4: Calculate the vertices of the polygon.
    vertex1 = closest_point + perpendicular1 * scale + perpendicular2 * scale
    vertex2 = closest_point - perpendicular1 * scale + perpendicular2 * scale
    vertex3 = closest_point - perpendicular1 * scale - perpendicular2 * scale
    vertex4 = closest_point + perpendicular1 * scale - perpendicular2 * scale
    # Step 5: Return the polygon.
    return np.array([vertex1, vertex2, vertex3, vertex4])



import mrdja.ransac.coreransac as coreransac
import mrdja.pointcloud as pointcloud
import mrdja.geometry as geometry
import mrdja.drawing as drawing
import open3d as o3d
import numpy as np

'''
Line-guided plane detection in pointclouds.
Tomar, por ejemplo, 200 rectas y luego los 40 mejores por parejas. Eso son 780 parejas. Tomar el 5% mejor según el error de ajuste del plano.
Eso son 39 parejas. En total 239 veces que hay que mirar los inliers mirando todos los puntos.
Comparar con las iteraciones de RANSAC que se necesitan para obtener el mismo o mayor número de inliers, un 5% menos, un 10% menos, etc.
También poner la diferencia entre los dos al llegar ambos al mismo número de iteraciones.
Comenzar con la misma semilla aleatoria para que las iteraciones de RANSAC sean las mismas.
'''

def partitions_with_respect_to_vector(reference_vector, list_vectors, angle_threshold):
    # returns a list of lists of vectors
    # list_vectors is traversed only once
    list_partitions = []
    for vector in list_vectors:
        # if the angle between the vector and all the vectors in some element of list_partitions is less than angle_threshold, 
        # then add the vector to that element; otherwise, create a new element with the reference_vector and vector

        # compute the angle between the vector and all the vectors in list_partitions
        for partition in list_partitions:
            print(partition)
            print(vector)
            all_angles = [geometry.get_angle_between_vector(vector, partition_vector) for partition_vector in partition]
            if all(abs(angle) < angle_threshold for angle in all_angles):
                partition.append(vector)
                break
        else:
            list_partitions.append([reference_vector, vector])
    return list_partitions

def partitions_with_respect_to_line(reference_line, list_lines, angle_threshold):
    # returns a list of lists of vectors
    # list_vectors is traversed only once
    list_partitions = []
    for index, line in enumerate(list_lines):
        # if the angle between the vector and all the vectors in some element of list_partitions is less than angle_threshold, 
        # then add the vector to that element; otherwise, create a new element with the reference_vector and vector

        # compute the angle between the vector and all the vectors in list_partitions
        for partition in list_partitions:
            print(partition)
            print(line)
            all_angles = [geometry.get_angle_between_lines(line, partition_line[0]) for partition_line in partition]
            if all(abs(angle) < angle_threshold for angle in all_angles):
                partition.append([line, index+1])
                break
        else:
            list_partitions.append([[reference_line, 0], [line, index+1]])
    return list_partitions

# create an example
reference_vector = np.array([1, 0, 0])
list_vectors = [np.array([1, 0, 0]), np.array([0, 1, 0]), np.array([0, 0, 1]), np.array([1, 1, 0]), np.array([1, 0, 1]), np.array([0, 1, 1]), np.array([1, 1, 1])]
angle_threshold = 0.1
# list_partitions = partitions_with_respect_to_vector(reference_vector, list_vectors, angle_threshold)

# The files in the Stanford database are like this:
# {'number_pcd_points': 1136617, 'has_normals': False, 
# 'has_colors': True, 'is_empty': False, 'max_x': -15.207, 
# 'min_x': -20.542, 'max_y': 41.283, 'min_y': 36.802, 
# 'max_z': 3.206, 'min_z': 0.02, 'all_points_finite': True, 
# 'all_points_unique': False}
# the measures are in meters
# dict_results = pointcloud_audit(pcd)
# print(dict_results)

def get_RANSAC_data_from_file(filename, ransac_iterations, threshold, seed):
    dict_full_results = {}
    dict_full_results["filename"] = filename

    pcd = o3d.io.read_point_cloud(filename)
    '''
    audit_before_sanitizing = pointcloud.pointcloud_audit(pcd)
    dict_full_results["audit_before_sanitizing"] = audit_before_sanitizing
    pcd = pointcloud.pointcloud_sanitize(pcd)
    audit_after_sanitizing = pointcloud.pointcloud_audit(pcd)
    dict_full_results["audit_after_sanitizing"] = audit_after_sanitizing
    '''
    number_pcd_points = len(pcd.points)
    dict_full_results["number_pcd_points"] = number_pcd_points
    np_points = np.asarray(pcd.points)

    dict_full_results["ransac_iterations_results"] = []

    np.random.seed(seed)
    max_number_inliers = 0
    best_iteration_results = None
    for _ in range(ransac_iterations):
        dict_iteration_results = coreransac.get_ransac_line_iteration_results(np_points, threshold, number_pcd_points)
        if dict_iteration_results["number_inliers"] > max_number_inliers:
            max_number_inliers = dict_iteration_results["number_inliers"]
            best_iteration_results = dict_iteration_results
        dict_full_results["ransac_iterations_results"].append(dict_iteration_results)
    dict_full_results["ransac_best_iteration_results"] = best_iteration_results
    return dict_full_results

ransac_iterations = 200
threshold = 0.02
seed = 42

filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/conferenceRoom_1/conferenceRoom_1.ply"
RANSAC_data_from_file = get_RANSAC_data_from_file(filename, ransac_iterations, threshold, seed)

print(RANSAC_data_from_file)

# create a function that takes the RANSAC_data_from_file and the iteration number and returns the inliers
# create a function that takes the RANSAC_data_from_file and the iteration number and returns the number of inliers

pcd = o3d.io.read_point_cloud(RANSAC_data_from_file["filename"])
np_points = np.asarray(pcd.points)

def get_inliers_from_iteration(RANSAC_data_from_file, np_points, iteration_number):
    dict_iteration_results = RANSAC_data_from_file["ransac_iterations_results"][iteration_number]
    inliers = np_points[dict_iteration_results["indices_inliers"]]
    pcd_inliers = o3d.geometry.PointCloud()
    pcd_inliers.points = o3d.utility.Vector3dVector(inliers)
    pcd_inliers.paint_uniform_color([1, 0, 0])
    return pcd_inliers

def get_number_inliers_from_iteration(RANSAC_data_from_file, iteration_number):
    dict_iteration_results = RANSAC_data_from_file["ransac_iterations_results"][iteration_number]
    return dict_iteration_results["number_inliers"]

# using open3d create a geometry with the inliers painted red, and show the inliers, the original point cloud and the number of iteration and the number
# of inliers in the same window; create a for loop that shows the inliers for each iteration

def draw_inliers_from_iteration(RANSAC_data_from_file, pcd, iteration_number):
    pcd_inliers = get_inliers_from_iteration(RANSAC_data_from_file, np_points, iteration_number)
    number_inliers = get_number_inliers_from_iteration(RANSAC_data_from_file, iteration_number)
    o3d.visualization.draw_geometries([pcd, pcd_inliers], window_name="Inliers from iteration " + str(iteration_number) + " with " + str(number_inliers) + " inliers")

for iteration_number in range(ransac_iterations):
    # only show if the number of inliers is greater than threshold
    threshold_number_inliers = 1000
    if get_number_inliers_from_iteration(RANSAC_data_from_file, iteration_number) > threshold_number_inliers:
        draw_inliers_from_iteration(RANSAC_data_from_file, pcd, iteration_number)

# create a function that takes the RANSAC_data_from_file and returns the number of inliers for each iteration

def get_number_inliers_from_RANSAC_data_from_file(RANSAC_data_from_file):
    number_inliers = []
    for dict_iteration_results in RANSAC_data_from_file["ransac_iterations_results"]:
        number_inliers.append(dict_iteration_results["number_inliers"])
    return number_inliers

number_inliers = get_number_inliers_from_RANSAC_data_from_file(RANSAC_data_from_file)

# create an histogram of the number of inliers for each iteration with matplotlib

import matplotlib.pyplot as plt

plt.hist(number_inliers, bins=100)
plt.show()

# create a function that returns all the current_line along with their number_inliers

def get_lines_and_number_inliers_from_RANSAC_data_from_file(RANSAC_data_from_file):
    pair_lines_number_inliers = []
    for dict_iteration_results in RANSAC_data_from_file["ransac_iterations_results"]:
        pair_lines_number_inliers.append((dict_iteration_results["current_line"], dict_iteration_results["number_inliers"]))
    return pair_lines_number_inliers

# order the pairs by number_inliers

def get_lines_and_number_inliers_ordered_by_number_inliers(RANSAC_data_from_file):
    pair_lines_number_inliers = get_lines_and_number_inliers_from_RANSAC_data_from_file(RANSAC_data_from_file)
    pair_lines_number_inliers_ordered = sorted(pair_lines_number_inliers, key=lambda pair_line_number_inliers: pair_line_number_inliers[1], reverse=True)
    return pair_lines_number_inliers_ordered

ordered_pairs = get_lines_and_number_inliers_ordered_by_number_inliers(RANSAC_data_from_file)

# get the first 50 pairs and get all the pairs of pairs

how_many_maximum = 0
list_sse_inliers = []
for i in range(40):
    for j in range(i+1, 40):
        line_1 = ordered_pairs[i][0]
        line_2 = ordered_pairs[j][0]
        points = np.array([line_1[0], line_1[1], line_2[0], line_2[1]])
        a, b, c, d, sse = geometry.fit_plane_svd(points)
        new_plane = np.array([a, b, c, d])
        threshold_pcd = 0.02
        how_many, inliers = coreransac.get_how_many_below_threshold_between_plane_and_points_and_their_indices(np_points, new_plane, threshold_pcd)
        list_sse_inliers.append((sse, how_many))
        if how_many > how_many_maximum:
            how_many_maximum = how_many
        if how_many > 200000:
            inliers = np.asarray(inliers)
            inlier_cloud = pcd.select_by_index(inliers)
            inlier_cloud.paint_uniform_color([1, 0, 0])
            o3d.visualization.draw_geometries([pcd, inlier_cloud], window_name="Inliers from iteration " + str(i) + " and " + str(j) + " with " + str(how_many) + " inliers")
print (how_many_maximum)

#plot the relationship between sse and number of inliers

list_sse_inliers = sorted(list_sse_inliers, key=lambda pair_sse_inliers: pair_sse_inliers[0])
list_sse = [pair_sse_inliers[0] for pair_sse_inliers in list_sse_inliers]
list_inliers = [pair_sse_inliers[1] for pair_sse_inliers in list_sse_inliers]
plt.plot(list_sse, list_inliers)
plt.show()

# plot a histogram of the sse

list_sse = [pair_sse_inliers[0] for pair_sse_inliers in list_sse_inliers]
plt.hist(list_sse, bins=100)
plt.show()

# get the elements of list_sse below the percentil 0.05 of list_sse

percentil = 0.05
list_sse = [pair_sse_inliers[0] for pair_sse_inliers in list_sse_inliers]
list_sse = sorted(list_sse)
print (list_sse[: int(len(list_sse)*percentil)])



# count how_many sse are below 0.05, 0.1, 0.2, 0.3, 0.4, 0.5

threshold_sse = 0.2
how_many_below_threshold = 0
for sse in list_sse:
    if sse < threshold_sse:
        how_many_below_threshold += 1
print (how_many_below_threshold)

# plot a histogram of the number of inliers

list_inliers = [pair_sse_inliers[1] for pair_sse_inliers in list_sse_inliers]
plt.hist(list_inliers, bins=100)
plt.show()

# plot a scatter plot of the sse and the number of inliers restricted to the sse values between 0 and 0.1

list_sse = [pair_sse_inliers[0] for pair_sse_inliers in list_sse_inliers]
list_inliers = [pair_sse_inliers[1] for pair_sse_inliers in list_sse_inliers]
plt.scatter(list_sse, list_inliers)
plt.xlim(0, 0.1)
plt.ylim(0, 100000)
plt.show()

# plot a scatter plot of the sse and the number of inliers restricted to the sse values below the 0.05 percentile

list_sse = [pair_sse_inliers[0] for pair_sse_inliers in list_sse_inliers]
list_inliers = [pair_sse_inliers[1] for pair_sse_inliers in list_sse_inliers]
percentile_05 = np.percentile(list_sse, 5)
list_sse_05 = [list_sse[i] for i in range(len(list_sse)) if list_sse[i] < percentile_05]
list_inliers_05 = [list_inliers[i] for i in range(len(list_sse)) if list_sse[i] < percentile_05]
plt.scatter(list_sse_05, list_inliers_05)
plt.xlim(0, 0.0002)
plt.ylim(0, 300000)
plt.show()

# plot a boxplot of the sse and the number of inliers restricted to the sse values between 0 and 0.1

list_sse = [pair_sse_inliers[0] for pair_sse_inliers in list_sse_inliers]
list_inliers = [pair_sse_inliers[1] for pair_sse_inliers in list_sse_inliers]
plt.boxplot(list_sse)
plt.ylim(-1, 10)
plt.show()

# get the sse values for the 95% percentile of the number of inliers

list_sse = [pair_sse_inliers[0] for pair_sse_inliers in list_sse_inliers]
list_inliers = [pair_sse_inliers[1] for pair_sse_inliers in list_sse_inliers]
percentile_95 = np.percentile(list_inliers, 95)
list_sse_95 = [list_sse[i] for i in range(len(list_sse)) if list_inliers[i] > percentile_95]
list_inliers_95 = [list_inliers[i] for i in range(len(list_inliers)) if list_inliers[i] > percentile_95]
print (list_sse_95)
print (list_inliers_95)

'''
angle_threshold = 0.3
how_many_maximum = 0
ordered_lines = [ordered_pair[0] for ordered_pair in ordered_pairs]
# ordered_vectors = [ordered_line[1] - ordered_line[0] for ordered_line in ordered_lines]
while len(ordered_lines) > 1:
    reference_line = ordered_lines[0]
    partitions = partitions_with_respect_to_line(reference_line, ordered_lines[1:], angle_threshold)
    print("Partitions: " + str(len(partitions)))
    print (partitions)
    # get the partition with the most elements
    partition_with_most_elements = max(partitions, key=lambda partition: len(partition))
    # get the lines and the indices from partition_with_most_elements
    lines = [line for line, index in partition_with_most_elements]
    indices = [index for line, index in partition_with_most_elements]
    # get the points of the partition with the most elements
    points = []
    for line in lines:
        points.append(line[0])
        points.append(line[1])
    points = np.asarray(points)
    # remove the partition with the most elements from the list of ordered_lines
    # sort the indices in descending order
    indices = sorted(indices, reverse=True)
    for index in indices:
        ordered_lines.pop(index)

    # get the plane that fits the points
    new_plane = geometry.fit_plane_svd(points)
    # get the number of inliers
    threshold_pcd = 0.02
    how_many, inliers = coreransac.get_how_many_below_threshold_between_plane_and_points_and_their_indices(np_points, new_plane, threshold_pcd)
    if how_many > how_many_maximum:
        how_many_maximum = how_many
    inliers = np.asarray(inliers)
    inlier_cloud = pcd.select_by_index(inliers)
    inlier_cloud.paint_uniform_color([1, 0, 0])
    o3d.visualization.draw_geometries([pcd, inlier_cloud], window_name="Inliers with " + str(how_many) + " inliers")
print (how_many_maximum)
'''



import mrdja.sampling as sampling
import matplotlib.pyplot as plt
import numpy as np
import mrdja.geometry as geometry
from mpl_toolkits.mplot3d import Axes3D

n_samples = 100
normal1 = (1, 1, 0)
normal2 = (0, 1, 0)
normal3 = (-1, 0, 1)
center = (1, 0, 1)
length1 = 5
length2 = 5
length3 = 5
normal1 = np.array(normal1)
normal2 = np.array(normal2)
normal3 = np.array(normal3)
center = np.array(center)
samples = sampling.sampling_parallelogram_3d(n_samples, normal1, normal2, normal3, center, length1, length2, length3)
samples = np.array(samples)

fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.scatter(samples[:, 0], samples[:, 1], samples[:, 2])

# create title from n_samples, center, and radius, using fstring
ax.set_title(f'{n_samples} Samples on a Parallelogram with normal vectors {normal1}, {normal2} and {normal3}, \
            center {center}, length1 {length1}, length2 {length2} and length3 {length3}')

# Draw the parallelogram
vertices = geometry.get_parallelogram_3d_vertices(center, normal1, normal2, normal3, length1, length2, length3)
# Define the edges of the 3d parallelogram
edges = [(0, 1), (1, 2), (2, 3), (3, 0), (4, 5), (5, 6),
         (6, 7), (7, 4), (0, 4), (1, 5), (2, 6), (3, 7)]
# Plot the edges
for edge in edges:
    ax.plot([vertices[edge[0]][0], vertices[edge[1]][0]],
            [vertices[edge[0]][1], vertices[edge[1]][1]],
            [vertices[edge[0]][2], vertices[edge[1]][2]], color='red')
    
# Draw the normals at a quarter of their corresponding length
quarter_length1 = length1 / 4
quarter_length2 = length2 / 4
arrow_length1 = quarter_length1 / 2
arrow_length2 = quarter_length2 / 2
ax.quiver(center[0], center[1], center[2], normal1[0], normal1[1], normal1[2], length=arrow_length1, normalize=False, color='red')
ax.quiver(center[0], center[1], center[2], normal2[0], normal2[1], normal2[2], length=arrow_length2, normalize=False, color='red')
ax.quiver(center[0], center[1], center[2], normal3[0], normal3[1], normal3[2], length=arrow_length2, normalize=False, color='red')

ax.set_xlabel('X')
ax.set_ylabel('Y')
ax.set_zlabel('Z')
plt.show()



import numpy as np
import open3d as o3d
from time import time
import math
from typing import Tuple, List
from numba import cuda


@cuda.jit
def get_how_many_below_threshold_kernel(points_x: np.ndarray, points_y: np.ndarray, points_z: np.ndarray,
                                         a: float, b: float, c: float, d: float,
                                         optimized_threshold: float, result: int) -> None:
    i = cuda.grid(1) # to compute the index of the current thread
    if i < points_x.shape[0]:
        dist = math.fabs(a * points_x[i] + b * points_y[i] + c * points_z[i] + d)
        if dist <= optimized_threshold:
            cuda.atomic.add(result, 0, 1)

filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/WC_1/WC_1.ply"
# filename = "/home/bee/S3DIS/Stanford3dDataset_v1.2/Area_1/WC_1/WC_1.ply"
pcd = o3d.io.read_point_cloud(filename)
np_points = np.asarray(pcd.points)
threshold_pcd = 22.52

# Extract x, y, z coordinates from np_points
points_x = np_points[:, 0]
points_y = np_points[:, 1]
points_z = np_points[:, 2]

# Convert points_x, points_y, and points_z to contiguous arrays
d_points_x = np.ascontiguousarray(points_x)
d_points_y = np.ascontiguousarray(points_y)
d_points_z = np.ascontiguousarray(points_z)

d_points_x = cuda.to_device(d_points_x)
d_points_y = cuda.to_device(d_points_y)
d_points_z = cuda.to_device(d_points_z)

num_points = np_points.shape[0]
iterations = 10

threadsperblock = 1024
max_threads_per_block = cuda.get_current_device().MAX_THREADS_PER_BLOCK
threadsperblock = min(max_threads_per_block, threadsperblock)
blockspergrid = math.ceil(num_points / threadsperblock)

t1 = time()
np.random.seed(42)
for _ in range(iterations):
    a, b, c, d = np.random.rand(4)
    # a, b, c, d = 0.1, 0.2, 0.3, 0.4
    new_plane = (a, b, c, d)
    # Output variable to store the result
    result = np.array([0], dtype=np.int32)
    d_result = cuda.to_device(result)
    optimized_threshold = threshold_pcd * math.sqrt(a * a + b * b + c * c)
    get_how_many_below_threshold_kernel[blockspergrid, threadsperblock](d_points_x, d_points_y, d_points_z, a, b, c, d, optimized_threshold, d_result)
    # Copy the result back to the host
    cuda.synchronize()
    result = d_result.copy_to_host()[0]
    print(result)

t2 =time()
print(f't2 - t1: {(t2-t1):.4f}s')

t1 = time()
plane_model, inliers = pcd.segment_plane(distance_threshold=threshold_pcd,
                                         ransac_n=3,
                                         num_iterations=1000)
t2 =time()
print(f't2 - t1: {(t2-t1):.4f}s')

t1 = time()
plane_model, inliers = pcd.segment_plane(distance_threshold=threshold_pcd,
                                         ransac_n=3,
                                         num_iterations=100000)
t2 =time()
print(f't2 - t1: {(t2-t1):.4f}s')



from mrdja.clustering import clustering_report
from sklearn.datasets import make_blobs
import matplotlib.pyplot as plt
import numpy as np

# Generate n-dimensional sample data using make_blobs
n_samples = 100  # Number of data points
n_features = 300    # Number of dimensions
centers = 3       # Number of clusters
X, y = make_blobs(n_samples=n_samples, n_features=n_features, centers=centers, random_state=42)

# perform PCA to reduce the dimensionality of the data
# from sklearn.decomposition import PCA
# pca = PCA(n_components=2)
# X = pca.fit_transform(X)

# Perform clustering with K-means, using some criteria to choose the right number of clusters
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
scores = []
max_clusters = 15
for k in range(2, max_clusters):
    clustering = KMeans(n_clusters=k)
    clustering.fit(X)
    score = silhouette_score(X, clustering.labels_)
    scores.append(score)
plt.plot(range(2, max_clusters), scores)
plt.xlabel("Number of clusters")
plt.ylabel("Silhouette score")
plt.show()
# choose the number of clusters that maximizes the silhouette score
# compute the number of clusters that maximizes the silhouette score
n_clusters = np.argmax(scores) + 2
clustering = KMeans(n_clusters=n_clusters)
clustering.fit(X)

# perform PCA to reduce the dimensionality of the data
from sklearn.decomposition import PCA
pca = PCA(n_components=2)
X = pca.fit_transform(X)
# plot the data, writing the percent of variance explained by each principal component
plt.scatter(X[:,0], X[:,1], c=clustering.labels_)
plt.xlabel("PC1 ({}%)".format(round(pca.explained_variance_ratio_[0]*100, 2)))
plt.ylabel("PC2 ({}%)".format(round(pca.explained_variance_ratio_[1]*100, 2)))
plt.show()

result, report = clustering_report(X)



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

a1 = 1.086
a2 = 3.084
a3 = 2.118
b1 = -1.274
b2 = -0.005
b3 = 0.198

# Define vectors A and B
A = np.array([a1, a2, a3])  # Replace with your values
B = np.array([b1, b2, b3])  # Replace with your values

# Create a meshgrid for plotting the plane defined by A and B
x = np.linspace(-10, 10, 10)
y = np.linspace(-10, 10, 10)
X, Y = np.meshgrid(x, y)

# Calculate the normal of the plane
normal = np.cross(A, B)

# Find a vector C perpendicular to A and inside the plane defined by A and B
C = np.cross(normal, A)

# normalize A, B and C
A = A / np.linalg.norm(A)
B = B / np.linalg.norm(B)
C = C / np.linalg.norm(C)

# Plot A, B , D and the plane
fig = plt.figure()
ax = fig.gca(projection='3d')
ax.quiver(0, 0, 0, A[0], A[1], A[2], color='g')
ax.quiver(0, 0, 0, B[0], B[1], B[2], color='g')
ax.quiver(0, 0, 0, C[0], C[1], C[2], color='b')
ax.quiver(0, 0, 0, normal[0], normal[1], normal[2], color='r')
ax.plot_surface(X, Y, (normal[0] * X + normal[1] * Y) / (-normal[2]), alpha=0.5)
ax.set_xlabel('X')
ax.set_ylabel('Y')
ax.set_zlabel('Z')
plt.show()







import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

# Generate random points in 3D space
num_points = 100
np.random.seed(42)
points = np.random.rand(num_points, 3)

# Define the function for fitting a plane using SVD
def fit_plane(points):
    centroid = np.mean(points, axis=0)
    shifted_points = points - centroid
    _, _, V = np.linalg.svd(shifted_points)
    normal = V[2]
    d = -np.dot(normal, centroid)
    return normal, d

# Fit the plane to the points
normal, d = fit_plane(points)

# Create a grid of points to represent the plane
xx, yy = np.meshgrid(np.linspace(-1, 2, 10), np.linspace(-1, 2, 10))
zz = (-normal[0] * xx - normal[1] * yy - d) / normal[2]

# Plot the points and the fitted plane
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.scatter(points[:, 0], points[:, 1], points[:, 2], c='b', marker='o')
ax.plot_surface(xx, yy, zz, alpha=0.2)

# Set labels and show the plot
ax.set_xlabel('X')
ax.set_ylabel('Y')
ax.set_zlabel('Z')
plt.show()



import numpy as np
import open3d as o3d
from time import time
import math
from typing import Tuple, List
from numba import cuda
import mrdja.ransac.coreransaccuda as coreransaccuda
import mrdja.ransac.coreransac as coreransac


@cuda.jit
def get_how_many_below_threshold_kernel(points_x: np.ndarray, points_y: np.ndarray, points_z: np.ndarray,
                                         a: float, b: float, c: float, d: float,
                                         optimized_threshold: float, result: int) -> None:
    i = cuda.grid(1) # to compute the index of the current thread
    if i < points_x.shape[0]:
        dist = math.fabs(a * points_x[i] + b * points_y[i] + c * points_z[i] + d)
        if dist <= optimized_threshold:
            cuda.atomic.add(result, 0, 1)

# filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/WC_1/WC_1.ply"
filename = "/tmp/Stanford3dDataset_v1.2/Area_1/WC_1/WC_1.ply"
pcd = o3d.io.read_point_cloud(filename)
np_points = np.asarray(pcd.points)
threshold_pcd = 22.52

# Extract x, y, z coordinates from np_points
points_x = np_points[:, 0]
points_y = np_points[:, 1]
points_z = np_points[:, 2]

# Convert points_x, points_y, and points_z to contiguous arrays
d_points_x = np.ascontiguousarray(points_x)
d_points_y = np.ascontiguousarray(points_y)
d_points_z = np.ascontiguousarray(points_z)

d_points_x = cuda.to_device(d_points_x)
d_points_y = cuda.to_device(d_points_y)
d_points_z = cuda.to_device(d_points_z)

num_points = np_points.shape[0]
iterations = 10

threadsperblock = 1024
max_threads_per_block = cuda.get_current_device().MAX_THREADS_PER_BLOCK
threadsperblock = min(max_threads_per_block, threadsperblock)
blockspergrid = math.ceil(num_points / threadsperblock)

t1 = time()
np.random.seed(42)
print("CUDA implementation in file")
for _ in range(iterations):
    a, b, c, d = np.random.rand(4)
    # a, b, c, d = 0.1, 0.2, 0.3, 0.4
    new_plane = (a, b, c, d)

    # Output variable to store the result
    result = np.array([0], dtype=np.int32)
    d_result = cuda.to_device(result)

    optimized_threshold = threshold_pcd * math.sqrt(a * a + b * b + c * c)
    get_how_many_below_threshold_kernel[blockspergrid, threadsperblock](d_points_x, d_points_y, d_points_z, a, b, c, d, optimized_threshold, d_result)

    # Copy the result back to the host
    cuda.synchronize()
    result = d_result.copy_to_host()[0]
    print(result)

t2 =time()
print(f't2 - t1: {(t2-t1):.4f}s')

t1 = time()
np.random.seed(42)
print("CUDA implementation (plane) in coreraansaccuda.py")
for _ in range(iterations):
    a, b, c, d = np.random.rand(4)
    # a, b, c, d = 0.1, 0.2, 0.3, 0.4
    new_plane = (a, b, c, d)
    # Output variable to store the result
    result = np.array([0], dtype=np.int32)
    d_result = cuda.to_device(result)
    other_result = coreransaccuda.get_how_many_below_threshold_between_plane_and_points_cuda(np_points, d_points_x, d_points_y, d_points_z, new_plane, threshold_pcd)
    print(other_result)
t2 =time()
print(f't2 - t1: {(t2-t1):.4f}s')

t1 = time()
np.random.seed(42)
print("CPU implementation (plane) in coreraansac.py")
for _ in range(iterations):
    a, b, c, d = np.random.rand(4)
    # a, b, c, d = 0.1, 0.2, 0.3, 0.4
    new_plane = (a, b, c, d)
    how_many, indices = coreransac.get_how_many_below_threshold_between_plane_and_points_and_their_indices(np_points, new_plane, threshold_pcd)
    print(how_many)
t2 =time()
print(f't2 - t1: {(t2-t1):.4f}s')

t1 = time()
np.random.seed(42)
print("CUDA implementation (line) in coreraansaccuda.py")
for _ in range(iterations):
    a, b, c, d, e, f = np.random.rand(6)
    # a, b, c, d = 0.1, 0.2, 0.3, 0.4
    line_two_points = ((a, b, c), (d, e, f))
    other_result = coreransaccuda.get_how_many_below_threshold_between_line_and_points_cuda(np_points, d_points_x, d_points_y, d_points_z, line_two_points, threshold_pcd)
    print(other_result)
t2 =time()
print(f't2 - t1: {(t2-t1):.4f}s')

t1 = time()
np.random.seed(42)
print("CPU implementation (line) in coreraansac.py")
for _ in range(iterations):
    a, b, c, d, e, f = np.random.rand(6)
    # a, b, c, d = 0.1, 0.2, 0.3, 0.4
    line_two_points = np.array([[a, b, c], [d, e, f]])
    how_many, indices = coreransac.get_how_many_below_threshold_between_line_and_points_and_their_indices(np_points, line_two_points, threshold_pcd)
    print(how_many)
t2 =time()
print(f't2 - t1: {(t2-t1):.4f}s')

print("Open3D implementation 1000")
t1 = time()
plane_model, inliers = pcd.segment_plane(distance_threshold=threshold_pcd,
                                         ransac_n=3,
                                         num_iterations=1000)
t2 =time()
print(f't2 - t1: {(t2-t1):.4f}s')

print("Open3D implementation 100000")
t1 = time()
plane_model, inliers = pcd.segment_plane(distance_threshold=threshold_pcd,
                                         ransac_n=3,
                                         num_iterations=100000)
t2 =time()
print(f't2 - t1: {(t2-t1):.4f}s')
